@@ -1,26 +1,27 @@
 # This file is part of Xpra.
-# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2017-2020 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import socket
 import struct
 import random
+
+from xpra.os_util import LINUX, monotonic_time, memoryview_to_bytes, hexstr
+from xpra.util import envint, ellipsizer
+from xpra.make_thread import start_thread
+from xpra.net.protocol import Protocol, READ_BUFFER_SIZE
+from xpra.net.bytestreams import SocketConnection, can_retry
+from xpra.log import Logger
+
 try:
     import errno
     EMSGSIZE = errno.EMSGSIZE
 except ImportError:
     EMSGSIZE = None
 
-from xpra.log import Logger
 log = Logger("network", "protocol", "udp")
 
-from xpra.os_util import LINUX, monotonic_time, memoryview_to_bytes
-from xpra.util import envint, repr_ellipsized
-from xpra.make_thread import start_thread
-from xpra.net.protocol import Protocol, READ_BUFFER_SIZE
-from xpra.net.bytestreams import SocketConnection, can_retry
-
 DROP_PCT = envint("XPRA_UDP_DROP_PCT", 0)
 DROP_FIRST = envint("XPRA_UDP_DROP_FIRST", 0)
 MIN_MTU = envint("XPRA_UDP_MIN_MTU", 576)
@@ -36,7 +37,11 @@ _header_struct = struct.Struct(b'!QQHHH')
 _header_size = _header_struct.size
 
 
-class PendingPacket(object):
+def h(v):
+    uuid_bin = struct.pack("!Q", v)
+    return hexstr(uuid_bin)
+
+class PendingPacket:
     def __init__(self, seqno, start_time, chunks=None):
         self.seqno = seqno
         self.start_time = start_time
@@ -44,10 +49,10 @@ class PendingPacket(object):
         self.chunk_gap = 0
         self.chunks = chunks
     def __repr__(self):
-        return ("PendingPacket(%i: %s chunks)" % (self.seqno, len(self.chunks or [])))
+        return "PendingPacket(%i: %s chunks)" % (self.seqno, len(self.chunks or b""))
 
 
-class UDPListener(object):
+class UDPListener:
     """
         This class is used by servers to receive UDP packets,
         it parses the header and then exposes the data received via process_packet_cb.
@@ -70,7 +75,7 @@ class UDPListener(object):
                 try:
                     buf, bfrom = self._socket.recvfrom(READ_BUFFER_SIZE)
                 except Exception as e:
-                    log("_read_thread_loop() buffer=%s, from=%s", repr_ellipsized(buf), bfrom, exc_info=True)
+                    log("_read_thread_loop() buffer=%s, from=%s", ellipsizer(buf), bfrom, exc_info=True)
                     if can_retry(e):
                         continue
                     raise
@@ -81,7 +86,7 @@ class UDPListener(object):
                 try:
                     self._process_packet_cb(self, *values)
                 except Exception as e:
-                    log("_read_thread_loop() buffer=%s, from=%s", repr_ellipsized(buf), bfrom, exc_info=True)
+                    log("_read_thread_loop() buffer=%s, from=%s", ellipsizer(buf), bfrom, exc_info=True)
                     if not self._closed:
                         log.error("Error: UDP packet processing error:")
                         log.error(" %s", e)
@@ -103,7 +108,7 @@ class UDPListener(object):
             try:
                 log("Protocol.close() calling %s", s.close)
                 s.close()
-            except:
+            except OSError:
                 log.error("error closing %s", s, exc_info=True)
             self._socket = None
         log("UDPListener.close() done")
@@ -120,7 +125,7 @@ class UDPProtocol(Protocol):
     """
 
     def __init__(self, *args, **kwargs):
-        Protocol.__init__(self, *args)
+        super().__init__(*args)
         self.mtu = 0
         self.last_sequence = -1     #the most recent packet sequence we processed in full
         self.highest_sequence = -1
@@ -174,17 +179,20 @@ class UDPProtocol(Protocol):
         if self._closed:
             return False
         missing = self._get_missing()
-        packet = ("udp-control", self.mtu, self.asynchronous_receive_enabled, self.last_sequence, self.highest_sequence, missing, tuple(self.cancel))
-        log("send_control() packet(%s)=%s", self.pending_packets, packet)
-        def send_control_failed():
-            #resend a new one
-            self.cancel_control_timer()
-            self.send_control()
-        self._send_async(packet, False, send_control_failed)
+        packet = ("udp-control", self.mtu, self.asynchronous_receive_enabled,
+                  self.last_sequence, self.highest_sequence, missing, tuple(self.cancel))
+        log("send_control() packet(%s)=%s", self.pending_packets, ellipsizer(packet))
+        self._add_packet_to_queue(packet, fail_cb=self.send_control_failed, synchronous=False)
         self.cancel = set()
         self.schedule_control()
         return False
 
+    def send_control_failed(self):
+        log("send_control_failed()")
+        #resend a new one
+        self.cancel_control_timer()
+        self.send_control()
+
     def _get_missing(self):
         """ the packets and chunks we are missing """
         if not self.pending_packets:
@@ -192,7 +200,7 @@ class UDPProtocol(Protocol):
         now = monotonic_time()
         max_time = now-self.jitter/1000.0
         missing = {}
-        for seqno, ip in self.pending_packets.items():
+        for seqno, ip in tuple(self.pending_packets.items()):
             start = ip.start_time
             if start>=max_time:
                 continue        #too recent, may still arrive
@@ -224,7 +232,8 @@ class UDPProtocol(Protocol):
         return missing
 
     def process_control(self, mtu, remote_async_receive, last_seq, high_seq, missing, cancel):
-        log("process_control(%i, %i, %i, %i, %s, %s) current seq=%i", mtu, remote_async_receive, last_seq, high_seq, missing, cancel, self.output_packetcount)
+        log("process_control(%i, %i, %i, %i, %s, %s) current seq=%i",
+            mtu, remote_async_receive, last_seq, high_seq, ellipsizer(missing), cancel, self.output_packetcount)
         con = self._conn
         if not con:
             return
@@ -234,17 +243,17 @@ class UDPProtocol(Protocol):
         #first, we can free all the packets that have been processed by the other end:
         #(resend cache and fail callback)
         if last_seq>=0:
-            done = [x for x in self.fail_cb.keys() if x<=last_seq]
+            done = tuple(x for x in tuple(self.fail_cb.keys()) if x<=last_seq)
             for x in done:
                 try:
                     del self.fail_cb[x]
-                except:
+                except KeyError:
                     pass
-            done = [x for x in self.resend_cache.keys() if x<=last_seq]
+            done = tuple(x for x in tuple(self.resend_cache.keys()) if x<=last_seq)
             for x in done:
                 try:
                     del self.resend_cache[x]
-                except:
+                except KeyError:
                     pass
         #next we can forget about sequence numbers that have been cancelled:
         #we don't need to request a re-send, and we can skip over them:
@@ -254,7 +263,7 @@ class UDPProtocol(Protocol):
                     self.can_skip.add(seqno)
                 try:
                     del self.pending_packets[seqno]
-                except:
+                except KeyError:
                     pass
             #we may now be able to move forward a bit:
             if self.pending_packets and (self.last_sequence+1) in self.can_skip:
@@ -268,12 +277,13 @@ class UDPProtocol(Protocol):
                 #hope for the best, and tell the other end to stop asking:
                 self.cancel.add(seqno)
                 continue
-            if len(missing_chunks)==0:
+            if not missing_chunks:
                 #the other end only knows it is missing the seqno,
                 #not how many chunks are missing, so send them all
                 missing_chunks = resend_cache.keys()
             if fail_cb_seq:
-                log("fail_cb[%i]=%s, missing_chunks=%s, len(resend_cache)=%i", seqno, repr_ellipsized(str(fail_cb_seq)), missing_chunks, len(resend_cache))
+                log("fail_cb[%i]=%s, missing_chunks=%s, len(resend_cache)=%i",
+                    seqno, ellipsizer(fail_cb_seq), missing_chunks, len(resend_cache))
                 #we have a fail callback for this packet,
                 #we have to decide if we send the missing chunks or use the callback,
                 #resend if the other end is missing less than 25% of the chunks:
@@ -282,18 +292,19 @@ class UDPProtocol(Protocol):
                     #too many are missing, forget about it
                     try:
                         del self.resend_cache[seqno]
-                    except:
+                    except KeyError:
                         pass
                     try:
                         del self.fail_cb[seqno]
-                    except:
+                    except KeyError:
                         pass
                     self.cancel.add(seqno)
                     fail_cb_seq()
                     continue
+            log("resending: %s", ellipsizer(missing_chunks))
             for c in missing_chunks:
                 data = resend_cache.get(c)
-                log("resend data[%i][%i]=%s", seqno, c, repr_ellipsized(str(data)))
+                #log("resend data[%i][%i]=%s", seqno, c, ellipsizer(data))
                 if data is None:
                     log.error("Error: cannot resend chunk %i of packet sequence %i", c, seqno)
                     log.error(" data missing from packet resend cache")
@@ -315,8 +326,9 @@ class UDPProtocol(Protocol):
             * otherwise queue it up and keep track of any missing sequence numbers,
               schedule a udp-control packet to notify the other end of what we're missing
         """
-        #log("process_udp_data%s %i bytes", (uuid, seqno, synchronous, chunk, chunks, repr_ellipsized(data), bfrom), len(data))
-        assert uuid==self.uuid
+        #log("process_udp_data%s %i bytes",
+        #    (uuid, seqno, synchronous, chunk, chunks, ellipsizer(data), bfrom), len(data))
+        assert uuid==self.uuid, "unexpected uuid in header, should be %s but received %s" % (h(self.uuid), h(uuid))
         if seqno<=self.last_sequence:
             log("skipping duplicate packet %5i.%i", seqno, chunk)
             return
@@ -331,7 +343,7 @@ class UDPProtocol(Protocol):
                 return
         self.highest_sequence = max(self.highest_sequence, seqno)
         if self.pending_packets or (synchronous and seqno!=self.last_sequence+1) or chunk!=0 or chunks!=1:
-            assert chunk>=0 and chunks>0 and chunk<chunks, "invalid chunk: %i/%i" % (chunk, chunks)
+            assert 0<=chunk<chunks, "invalid chunk: %i/%i" % (chunk, chunks)
             #slow path: add chunk to incomplete packet
             now = monotonic_time()
             ip = self.pending_packets.get(seqno)
@@ -342,28 +354,44 @@ class UDPProtocol(Protocol):
                 self.pending_packets[seqno] = ip
             else:
                 ip.last_time = now
+            log("process_udp_data: sequence %i, got chunk %i/%i (%i bytes)", seqno, chunk+1, chunks, len(data))
             ip.chunks[chunk] = data
-            if seqno!=self.last_sequence+1:
+            if seqno>self.last_sequence+1:
                 #we're waiting for a packet and this is not it,
                 #make sure any gaps are marked as incomplete:
                 for i in range(self.last_sequence+1, seqno):
                     if i not in self.pending_packets and i not in self.can_skip:
                         self.pending_packets[i] = PendingPacket(i, now)
                 #make sure we request the missing packets:
-                self.schedule_control(self.jitter)
+                mcount = seqno-self.last_sequence
+                self.schedule_control(self.jitter*2//mcount)
                 if synchronous:
                     #we have to wait for the missing chunks / packets
                     log("process_udp_data: queuing %i as we're still waiting for %i", seqno, self.last_sequence+1)
                     return
-            if any(x is None for x in ip.chunks):
-                #one of the chunks is still missing
-                log("process_udp_data: sequence %i, got chunk %i but still missing: %s", seqno, chunk, [i for i,x in enumerate(ip.chunks) if x is None])
-                self.schedule_control(self.jitter)
+            missing = tuple(i for i, x in enumerate(ip.chunks) if x is None)
+            if missing:
+                #some chunks are missing
+                log("process_udp_data: total received: %i bytes, still missing %s chunks",
+                    sum(len(x or b"") for x in ip.chunks), ellipsizer(missing))
+                if min(missing)<chunk:
+                    #out of order, so more likely to really be missing
+                    delay = max(10, min(len(ip.chunks)//10, 100))*self.jitter//10
+                else:
+                    #still in correct order, so likely to arrive,
+                    #wait longer before sending the control packet:
+                    now = monotonic_time()
+                    elapsed = now-ip.start_time
+                    lm = len(missing)
+                    received = len(ip.chunks)-lm
+                    delay = self.jitter+min(1000, int(lm*elapsed/received))
+                self.schedule_control(delay)
                 return
             #all the data is here!
             del self.pending_packets[seqno]
             data = b"".join(ip.chunks)
-        log("process_udp_data: adding packet sequence %5i to read queue (got final chunk %i, synchronous=%s)", seqno, chunk, synchronous!=0)
+        log("process_udp_data: packet sequence %5i : %i bytes added to read queue (got final chunk %i, synchronous=%s)",
+            seqno, len(data), chunk, synchronous!=0)
         if seqno==self.last_sequence+1:
             self.last_sequence = seqno
         else:
@@ -406,30 +434,11 @@ class UDPProtocol(Protocol):
             self.last_sequence = seqno
             self._read_queue_put(data)
 
-
-    def _send_async(self, packet, sync=False, fail_cb=None):
-        """ used by send_control to bypass the regular queuing functions,
-            and force enable synchronous=False
-        """
-        #log("_send_async(%s, %s) encoder=%s, compressor=%s", packet, sync, self._encoder, self._compress)
-        log("_send_async(%s, %s)", packet, sync)
-        chunks = self.encode(packet)
-        if len(chunks)>1:
-            return Protocol.send_now(self, packet)
-        proto_flags,index,level,data = chunks[0]
-        from xpra.net.header import pack_header
-        payload_size = len(data)
-        header_and_data = pack_header(proto_flags, level, index, payload_size) + data
-        with self._write_lock:
-            if self._write_thread is None:
-                self.start_write_thread()
-            self._write_queue.put((header_and_data, None, None, fail_cb, sync))
-
-    def raw_write(self, items, start_cb=None, end_cb=None, fail_cb=None, synchronous=True, _more=False):
+    def raw_write(self, packet_type, items, start_cb=None, end_cb=None, fail_cb=None, synchronous=True, _more=False):
         """ make sure we don't enable asynchronous mode until the other end is read """
-        if not self.asynchronous_send_enabled:
+        if packet_type!="udp-control" and not self.asynchronous_send_enabled:
             synchronous = True
-        Protocol.raw_write(self, items, start_cb, end_cb, fail_cb, synchronous)
+        super().raw_write(packet_type, items, start_cb, end_cb, fail_cb, synchronous)
 
     def write_buffers(self, buf_data, fail_cb, synchronous):
         """
@@ -461,7 +470,8 @@ class UDPProtocol(Protocol):
         chunks = l // maxpayload
         if l % maxpayload > 0:
             chunks += 1
-        log("UDP.write_buf(%s, %i bytes, %s, %s) seq=%i, mtu=%s, maxpayload=%i, chunks=%i, data=%s", con, l, fail_cb, synchronous, seqno, mtu, maxpayload, chunks, repr_ellipsized(data))
+        log("UDP.write_buf(%s, %6i bytes, %s, %5s) uuid=%s, seq=%-5i, mtu=%5s, maxpayload=%5i, chunks=%-5i, data=%s",
+            con, l, fail_cb, synchronous, h(self.uuid), seqno, mtu, maxpayload, chunks, ellipsizer(data))
         chunk = 0
         offset = 0
         if fail_cb:
@@ -487,7 +497,7 @@ class UDPProtocol(Protocol):
 
 
     def get_info(self, alias_info=True):
-        i = Protocol.get_info(self, alias_info)
+        i = super().get_info(alias_info)
         i.update({
             "mtu"   : {
                 ""      : clamp_mtu(self.mtu),
@@ -508,11 +518,12 @@ class UDPServerProtocol(UDPProtocol):
 class UDPClientProtocol(UDPProtocol):
 
     def __init__(self, *args, **kwargs):
-        UDPProtocol.__init__(self, *args, uuid = random.randint(0, 2**64-1))
+        UDP_CLIENT_UUID = envint("XPRA_UDP_CLIENT_UUID", 0)
+        super().__init__(*args, uuid = UDP_CLIENT_UUID or random.randint(0, 2**64-1))
 
     def con_write(self, data, fail_cb):
         """ After successfully writing some data, update the mtu value """
-        r = UDPProtocol.con_write(self, data, fail_cb)
+        r = super().con_write(data, fail_cb)
         if r>0 and LINUX:
             IP_MTU = 14
             con = self._conn
@@ -529,8 +540,8 @@ class UDPClientProtocol(UDPProtocol):
             Splits and parses the UDP frame header from the packet,
             then process the packed using process_udp_data
         """
-        #log.info("UDPClientProtocol.read_queue_put(%s)", repr_ellipsized(buf))
         uuid, seqno, synchronous, chunk, chunks = _header_struct.unpack_from(buf[:_header_size])
+        #log("UDPClientProtocol: header uuid=%s", uuid)
         data = buf[_header_size:]
         bfrom = None        #not available here..
         self.process_udp_data(uuid, seqno, synchronous, chunk, chunks, data, bfrom)
@@ -544,23 +555,19 @@ class UDPSocketConnection(SocketConnection):
         they do not call connect() and so we have to specify the remote target every time)
     """
 
-    def __init__(self, *args):
-        SocketConnection.__init__(self, *args)
-
     def write(self, buf):
         #log("UDPSocketConnection: sending %i bytes to %s", len(buf), self.remote)
         try:
             return self._socket.sendto(buf, self.remote)
         except IOError as e:
             if e.errno==EMSGSIZE:
-                raise MTUExceeded("invalid UDP payload size, cannot send %i bytes: %s" % (len(buf), e))
+                raise MTUExceeded("invalid UDP payload size, cannot send %i bytes: %s" % (len(buf), e)) from None
             raise
 
     def close(self):
         """
-            don't close the socket, we're don't own it
+            don't close the socket, we don't own it
         """
-        pass
 
 class MTUExceeded(IOError):
     pass
